# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from ironic.common.i18n import _


CONF = cfg.CONF

opts = [
    cfg.StrOpt('temp_dir',
               help=_('When local RPC is used (rpc_transport=None), this is '
                      'the name of the directory to create temporary files '
                      'in. Must not be readable by any other processes. '
                      'If not provided, a temporary directory is used.')),
    cfg.BoolOpt('use_ssl',
                default=True,
                help=_('Whether to use TLS on the local RPC bus. Only set to '
                       'False if you experience issues with TLS and if all '
                       'local processes are trusted!')),
]


def register_opts(conf):
    conf.register_opts(opts, group='local_rpc')


def list_opts():
    return opts



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import os
import secrets
import tempfile

from keystoneauth1 import loading as ks_loading
from oslo_log import log

from ironic.common import auth_basic
from ironic.common.json_rpc import client
from ironic.common import tls_utils
from ironic.common import utils
from ironic.conf import CONF


LOG = log.getLogger(__name__)


_PASSWORD_BYTES = 64
_VALID_FOR_DAYS = 9999  # rotation not possible
_USERNAME = 'ironic'


def _lo_has_ipv6():
    return (
        os.path.exists("/proc/sys/net/ipv6/conf/lo/disable_ipv6")
        and open("/proc/sys/net/ipv6/conf/lo/disable_ipv6").read() != "1"
    )


def _create_tls_files(ip):
    with tempfile.NamedTemporaryFile(
            delete=False, dir=CONF.local_rpc.temp_dir) as fp:
        cert_file = fp.name
    with tempfile.NamedTemporaryFile(
            delete=False, dir=CONF.local_rpc.temp_dir) as fp:
        key_file = fp.name

    tls_utils.generate_tls_certificate(cert_file, key_file,
                                       common_name='ironic',
                                       ip_address=ip,
                                       valid_for_days=_VALID_FOR_DAYS)
    return cert_file, key_file


def _create_htpasswd(password):
    with tempfile.NamedTemporaryFile(
            mode="w+t", delete=False, dir=CONF.local_rpc.temp_dir) as fp:
        auth_basic.write_password(fp, _USERNAME, password)
        return fp.name


def configure():
    """Configure the local JSON RPC bus (if enabled)."""
    if CONF.rpc_transport != 'none':
        return

    ip = '::1' if _lo_has_ipv6() else '127.0.0.1'
    LOG.debug('Configuring local RPC bus on %s:%d', ip, CONF.json_rpc.port)

    if CONF.local_rpc.use_ssl:
        cert_file, key_file = _create_tls_files(ip)

        def _cleanup():
            utils.unlink_without_raise(cert_file)
            utils.unlink_without_raise(key_file)

        atexit.register(_cleanup)
    else:
        cert_file, key_file = None, None
    password = secrets.token_urlsafe(_PASSWORD_BYTES)
    htpasswd_path = _create_htpasswd(password)

    # NOTE(dtantsur): it is not possible to override username/password  without
    # registering http_basic options first.
    opts = ks_loading.get_auth_plugin_conf_options('http_basic')
    CONF.register_opts(opts, group='json_rpc')

    for key, value in [
        ('use_ssl', CONF.local_rpc.use_ssl),
        # Client options
        ('auth_type', 'http_basic'),
        ('cafile', cert_file),
        ('username', _USERNAME),
        ('password', password),
        # Server options
        ('auth_strategy', 'http_basic'),
        ('http_basic_auth_user_file', htpasswd_path),
        ('host_ip', ip),
        ('cert_file', cert_file),
        ('key_file', key_file),
    ]:
        CONF.set_override(key, value, group='json_rpc')


class LocalClient(client.Client):
    """JSON RPC client that always connects to the server's host IP."""

    def prepare(self, topic, version=None):
        # TODO(dtantsur): check that topic matches the expected host name
        # (which is not host_ip, by the way, it's CONF.host).
        return self.prepare_for_target(CONF.json_rpc.host_ip)



# Copyright 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTE(dtantsur): partial copy from IPA commit
# d86923e7ff40c3ec1d43fe9d4068f0bd3b17de67

import datetime
import ipaddress

from cryptography.hazmat import backends
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography import x509
from oslo_log import log


LOG = log.getLogger(__name__)


def _create_private_key(output):
    """Create a new private key and write it to a file.

    Using elliptic curve keys since they are 2x smaller than RSA ones of
    the same security (the NIST P-256 curve we use roughly corresponds
    to RSA with 3072 bits).

    :param output: Output file name.
    :return: a private key object.
    """
    private_key = ec.generate_private_key(ec.SECP256R1(),
                                          backends.default_backend())
    pkey_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )
    with open(output, 'wb') as fp:
        fp.write(pkey_bytes)

    return private_key


def generate_tls_certificate(output, private_key_output,
                             common_name, ip_address,
                             valid_for_days=30):
    """Generate a self-signed TLS certificate.

    :param output: Output file name for the certificate.
    :param private_key_output: Output file name for the private key.
    :param common_name: Content for the common name field (e.g. host name).
    :param ip_address: IP address the certificate will be valid for.
    :param valid_for_days: Number of days the certificate will be valid for.
    :return: the generated certificate as a string.
    """
    if isinstance(ip_address, str):
        ip_address = ipaddress.ip_address(ip_address)

    private_key = _create_private_key(private_key_output)

    subject = x509.Name([
        x509.NameAttribute(x509.NameOID.COMMON_NAME, common_name),
    ])
    alt_name = x509.SubjectAlternativeName([x509.IPAddress(ip_address)])
    not_valid_before = datetime.datetime.now(tz=datetime.timezone.utc)
    not_valid_after = (datetime.datetime.now(tz=datetime.timezone.utc)
                       + datetime.timedelta(days=valid_for_days))
    cert = (x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(subject)
            .public_key(private_key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(not_valid_before)
            .not_valid_after(not_valid_after)
            .add_extension(alt_name, critical=True)
            .sign(private_key, hashes.SHA256(), backends.default_backend()))
    pub_bytes = cert.public_bytes(serialization.Encoding.PEM)
    with open(output, "wb") as f:
        f.write(pub_bytes)
    LOG.info('Generated TLS certificate for IP address %s valid from %s '
             'to %s', ip_address, not_valid_before, not_valid_after)
    return pub_bytes.decode('utf-8')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ipaddress
import os
from unittest import mock

import bcrypt
from cryptography.hazmat import backends
from cryptography import x509

from ironic.common import utils
from ironic.conductor import local_rpc
from ironic.conf import CONF
from ironic.tests import base as tests_base


@mock.patch('atexit.register', autospec=True)
@mock.patch.object(local_rpc, '_lo_has_ipv6', autospec=True)
class ConfigureTestCase(tests_base.TestCase):

    def setUp(self):
        super().setUp()
        CONF.set_override('rpc_transport', 'none')
        self.addCleanup(self._cleanup_files)

    def _cleanup_files(self):
        if CONF.json_rpc.cert_file:
            utils.unlink_without_raise(CONF.json_rpc.cert_file)
        if CONF.json_rpc.key_file:
            utils.unlink_without_raise(CONF.json_rpc.key_file)
        if CONF.json_rpc.http_basic_auth_user_file:
            utils.unlink_without_raise(CONF.json_rpc.http_basic_auth_user_file)

    def _verify_tls(self, ipv6=True):
        self.assertTrue(os.path.exists(CONF.json_rpc.key_file))
        with open(CONF.json_rpc.cert_file, 'rb') as fp:
            cert = x509.load_pem_x509_certificate(
                fp.read(), backends.default_backend())
        # NOTE(dtantsur): most of the TLS generation is tested in
        # test_tls_utils, here only the relevant parts
        subject_alt_name = cert.extensions.get_extension_for_oid(
            x509.ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
        expected = (ipaddress.IPv6Address('::1') if ipv6
                    else ipaddress.IPv4Address('127.0.0.1'))
        self.assertEqual(
            [expected],
            subject_alt_name.value.get_values_for_type(x509.IPAddress))

    def _verify_password(self):
        self.assertEqual('ironic', CONF.json_rpc.username)
        self.assertTrue(CONF.json_rpc.password)
        with open(CONF.json_rpc.http_basic_auth_user_file) as fp:
            username, hashed = fp.read().strip().split(':', 1)
        self.assertEqual(username, CONF.json_rpc.username)
        self.assertTrue(
            bcrypt.checkpw(CONF.json_rpc.password.encode(), hashed.encode()))

    def test_wrong_rpc_transport(self, mock_lo_has_ipv6, mock_atexit_register):
        CONF.set_override('rpc_transport', 'oslo')
        local_rpc.configure()
        mock_lo_has_ipv6.assert_not_called()
        mock_atexit_register.assert_not_called()
        self.assertIsNone(CONF.json_rpc.cert_file)

    def test_default(self, mock_lo_has_ipv6, mock_atexit_register):
        mock_lo_has_ipv6.return_value = True

        local_rpc.configure()

        self.assertTrue(CONF.json_rpc.use_ssl)
        self.assertEqual('http_basic', CONF.json_rpc.auth_type)
        self.assertEqual('http_basic', CONF.json_rpc.auth_strategy)
        self.assertEqual('::1', CONF.json_rpc.host_ip)
        self._verify_password()
        self._verify_tls(ipv6=True)

    def test_ipv4(self, mock_lo_has_ipv6, mock_atexit_register):
        mock_lo_has_ipv6.return_value = False

        local_rpc.configure()

        self.assertTrue(CONF.json_rpc.use_ssl)
        self.assertEqual('http_basic', CONF.json_rpc.auth_type)
        self.assertEqual('http_basic', CONF.json_rpc.auth_strategy)
        self.assertEqual('127.0.0.1', CONF.json_rpc.host_ip)
        self._verify_password()
        self._verify_tls(ipv6=False)
